import os
import time

import pytest
from nyckel import OAuth2Renewer, TextClassificationFunction, TextClassificationSample, ClassificationPrediction


@pytest.mark.skipif("NYCKEL_CLIENT_ID" not in os.environ, reason="Env. var NYCKEL_CLIENT_ID not set")
@pytest.mark.skipif("NYCKEL_CLIENT_SECRET" not in os.environ, reason="Env. var NYCKEL_CLIENT_SECRET not set")
def test_end_to_end():
    auth = OAuth2Renewer(
        client_id=os.environ["NYCKEL_CLIENT_ID"],
        client_secret=os.environ["NYCKEL_CLIENT_SECRET"],
        server_url="https://www.nyckel.com",
    )

    samples = [
        TextClassificationSample(data="hello", annotation="Nice"),
        TextClassificationSample(data="hi", annotation="Nice"),
        TextClassificationSample(data="Good bye", annotation="Boo"),
        TextClassificationSample(data="I'm leaving", annotation="Boo"),
        TextClassificationSample(data="Hi again"),
    ]

    rate_my_greeting_function = TextClassificationFunction.new("RateMyGreeting", auth)

    rate_my_greeting_function.add_labels(["Nice", "Boo"])
    rate_my_greeting_function.add_samples(samples)

    has_trained_model = False
    while not has_trained_model:
        time.sleep(1)
        has_trained_model = rate_my_greeting_function.has_trained_model()
        print("No trained model yet. Sleeping 1 sec.")

    assert isinstance(rate_my_greeting_function("Howdy"), ClassificationPrediction)

    assert len(rate_my_greeting_function.invoke(["Hej", "Hola", "Gruezi"])) == 3

    rate_my_greeting_function.delete()



from .functions.text_classification_function import (
    TextClassificationFunction,  # noqa: F401
    TextClassificationSample,  # noqa: F401
    ClassificationPrediction,  # noqa: F401
)
from .auth import OAuth2Renewer  # noqa: F401



import concurrent.futures
from typing import Dict, List

import requests
from tqdm import tqdm

from nyckel.config import NBR_CONCURRENT_REQUESTS


class ParallelPoster:
    def __init__(self, session: requests.Session, endpoint: str):
        self._session = session
        self._endpoint = endpoint

    def _post_as_json(self, data: Dict):
        response = self._session.post(self._endpoint, json=data)
        return response

    def __call__(self, bodies: List[Dict]) -> Dict[Dict, requests.Response]:
        responses = [None] * len(bodies)
        n_workers = min(len(bodies), NBR_CONCURRENT_REQUESTS)

        with concurrent.futures.ThreadPoolExecutor(max_workers=n_workers) as executor:
            index_by_future = {executor.submit(self._post_as_json, body): index for index, body in enumerate(bodies)}
            for future in tqdm(concurrent.futures.as_completed(index_by_future)):
                index = index_by_future[future]
                body = bodies[index]
                response = future.result()
                if response.status_code not in [200, 409]:
                    print(f"Posting {body} to {self._endpoint} failed with {response.status_code=} {response.text=}")
                else:
                    pass  # It worked.
                responses[index] = response
        return responses


def repeated_get(session: requests.Session, endpoint: str):
    base_url, slug = endpoint.split(".com")
    base_url += ".com"
    resp = session.get(base_url + slug)
    resource_list = resp.json()
    while "next" in resp.links:
        slug = resp.links["next"]["url"]
        response = session.get(base_url + slug)
        assert resp.status_code == 200, f"GET from {base_url+slug} failed with {response.status_code}, {response.text}."
        resource_list.extend(resp.json())

    return resource_list



from dataclasses import dataclass
from typing import List, Optional

import requests
from requests.adapters import HTTPAdapter, Retry

from nyckel.auth import OAuth2Renewer
from nyckel.functions.utils import strip_nyckel_prefix
from nyckel.request_utils import ParallelPoster, repeated_get

LabelName = str


@dataclass
class ClassificationPrediction:
    label_name: LabelName
    confidence: float


@dataclass
class TextClassificationSample:
    data: str
    annotation: Optional[LabelName] = None
    prediction: Optional[ClassificationPrediction] = None


class TextClassificationFunction:
    def __init__(self, function_id: str, auth: OAuth2Renewer):
        self._function_id = function_id
        self._auth = auth
        self._session = requests.Session()
        retries = Retry(total=5, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
        self._session.mount("https://", HTTPAdapter(max_retries=retries))

    @classmethod
    def new(cls, name: str, auth: OAuth2Renewer):
        session = requests.Session()
        session.headers.update({"authorization": "Bearer " + auth.token})
        response = session.post(
            f"{auth.server_url}/v1/functions/", json={"input": "Text", "output": "Classification", "name": name}
        )

        assert response.status_code == 200, f"Something went wrong when creating function: {response.text}"
        prefixed_function_id = response.json()["id"]
        function_id = strip_nyckel_prefix(prefixed_function_id)
        print(f"Created function {name} with id: {function_id}")
        return TextClassificationFunction(function_id, auth)

    def __str__(self):
        return self.train_page

    def __repr__(self):
        return self.train_page

    @property
    def train_page(self):
        return f"{self._auth.server_url}/console/functions/{self._function_id}/train"

    def has_trained_model(self) -> bool:
        self._refresh_auth_token()
        body = {"data": "dummy data"}
        endpoint = self.api_endpoint("invoke")
        response = self._session.post(endpoint, json=body)
        if "No model available" in response.text:
            return False
        if not response.status_code == 200:
            return False
        return True

    def api_endpoint(self, slug, api_version="v1"):
        if slug == "":
            return f"{self._auth.server_url}/{api_version}/functions/{self._function_id}"
        else:
            return f"{self._auth.server_url}/{api_version}/functions/{self._function_id}/{slug}"

    def _refresh_auth_token(self):
        self._session.headers.update({"authorization": "Bearer " + self._auth.token})

    def __call__(self, sample_data: str) -> ClassificationPrediction:
        self._refresh_auth_token()
        body = {"data": sample_data}
        endpoint = self.api_endpoint("invoke")
        response = self._session.post(endpoint, json=body)
        if "No model available" in response.text:
            print(f"No model trained yet for this function. Go to {self.train_page} to see function status.")
            return None

        assert response.status_code == 200, f"Invoke failed with {response.status_code=}, {response.text=}"
        return ClassificationPrediction(
            label_name=response.json()["labelName"], confidence=response.json()["confidence"]
        )

    def invoke(self, sample_data_list: List[str]) -> List[ClassificationPrediction]:
        self._refresh_auth_token()
        print(f"Invoking {len(sample_data_list)} labels to {self.train_page} ...")
        bodies = [{"data": sample_data} for sample_data in sample_data_list]
        endpoint = self.api_endpoint("invoke")
        response_list = ParallelPoster(self._session, endpoint)(bodies)
        return [
            ClassificationPrediction(label_name=response.json()["labelName"], confidence=response.json()["confidence"])
            for response in response_list
        ]

    def add_labels(self, label_names=List[str]):
        self._refresh_auth_token()
        print(f"Posting {len(label_names)} labels to {self.train_page} ...")
        bodies = [{"name": label_name} for label_name in label_names]
        endpoint = self.api_endpoint("labels")
        ParallelPoster(self._session, endpoint)(bodies)

    def add_samples(self, samples: List[TextClassificationSample]):
        self._refresh_auth_token()
        print(f"Posting {len(samples)} samples to {self.train_page} ...")
        bodies = [{"data": s.data, "annotation": {"labelName": s.annotation}} for s in samples if s.annotation]
        bodies.extend([{"data": s.data} for s in samples if not s.annotation])
        endpoint = self.api_endpoint("samples")
        ParallelPoster(self._session, endpoint)(bodies)

    def delete(self) -> None:
        self._refresh_auth_token()
        endpoint = self.api_endpoint("")
        response = self._session.delete(endpoint)
        assert response.status_code == 200, f"Delete failed with {response.status_code=}, {response.text=}"
        print(f"Function {self.train_page} deleted.")

    def get_samples(self) -> List[TextClassificationSample]:
        self._refresh_auth_token()
        samples_return = repeated_get(self._session, self.api_endpoint("samples"))
        return samples_return
